/*
 *********************************************************************

 $Id$


 Copyright (c) 2007-2014 Whitestein Technologies AG,
 Riedstrasse 13, CH-6330 Cham, Switzerland.
 All rights reserved.

 This software is confidential and proprietary information of
 Whitestein Technologies AG.
 You shall not disclose this confidential information and shall use
 it only in accordance with the terms of the license agreement you
 entered into with Whitestein Technologies AG.
 The use of this file in source or binary form requires a written
 license from Whitestein Technologies AG.
 *********************************************************************
 */
package com.whitestein.lsps.pm.ldap;

import java.io.IOException;
import java.io.InputStream;
import java.nio.charset.StandardCharsets;
import java.util.Properties;

import javax.annotation.PostConstruct;
import javax.annotation.PreDestroy;
import javax.annotation.security.PermitAll;
import javax.ejb.EJB;

import com.novell.ldap.LDAPConnection;
import com.novell.ldap.LDAPException;
import com.whitestein.lsps.orgstructure.dao.RoleDao;
import com.whitestein.lsps.orgstructure.dao.RoleNameDao;
import com.whitestein.lsps.orgstructure.dao.SecurityRoleDao;

/**
 * Base class for all LDAP connecting ejbs, contains logic of connecting to LDAP.
 *
 *
 * @author mbj
 *
 */
@PermitAll
public abstract class AbstractLdapBean {

	@EJB
	protected SecurityRoleDao securityRoleDao;

	@EJB
	protected RoleDao roleDao;

	@EJB
	protected RoleNameDao roleNameDao;

	/**
	 * LDAP connection.
	 */
	protected LDAPConnection connection;

	/**
	 * Settings for the LDAP connection and communication.
	 */
	protected LdapSettings settings;

	/**
	 * Load all configuration from classpath file "/ldap-pm.properties" and
	 * connects to the server. The following properties are accepted:
	 *
	 * <ul>
	 * <li>com.whitestein.ldap.host : host of ldap server </li>
	 * <li>com.whitestein.ldap.port : port on which the ldap server is running on</li>
	 * <li>com.whitestein.ldap.servicelogin : login to use when connecting to the server</li>
	 * <li>com.whitestein.ldap.servicepassword : password to use when connecting to the server</li>
	 * <li>com.whitestein.ldap.base : base of the ldap tree in which all information are stored
	 * 		(i.e. all user entries must be in a subtree of this base item</li>
	 * <li>com.whitestein.ldap.idAttribute : name of the attribute which corresponds to the person id</li>
	 * <li>com.whitestein.ldap.userClass : distinguishing class - all entries which are to be used in lsps must have this class</li>
	 * <li>com.whitestein.ldap.loginNameAttribute : name of the attribute which corresponds to the person login</li>
	 * <li>com.whitestein.ldap.firstNameAttribute : name of the attribute which corresponds to the given name</li>
	 * <li>com.whitestein.ldap.lastNameAttribute : name of the attribute which corresponds to the last name</li>
	 * <li>com.whitestein.ldap.emailAttribute : name of the attribute which corresponds to the email</li>
	 * <li>com.whitestein.ldap.phoneAttribute : name of the attribute which corresponds to the mobile phone</li>
	 * <li>com.whitestein.ldap.settingsAttribute :  : name of the attribute which corresponds to the lsps settings (settings stored by lsps in binary form)</li>
	 * <li>com.whitestein.ldap.passwordAttribute : name of the attribute which corresponds to the password attribute (used to update password of the user.
	 * 		To verify password separate logging in to the ldap with user credentials is used</li>
	 * <li>com.whitestein.ldap.propertiesAttributes : either * for all LDAP attributes, or list of attribute names delimited by comma - these attributes will be used as person details</li>
	 * <li>com.whitestein.ldap.securityRoleAttribute : name of the attribute which corresponds to the security role (under this attribute list of security roles is stored)</li>
	 * <li>com.whitestein.ldap.substitutionActive : name of the attribute which corresponds to the flag indicating active substitution for a person</li>
	 * <li>com.whitestein.ldap.substitutes : name of the attribute which stores persons deputies</li>
	 * <li>com.whitestein.ldap.readonly : if true, application will not enable editing user related data directly from LSPS</li>
	 * </ul>
	 *
	 */
	@PostConstruct
	public void connect() {
		// init configuration
		Properties properties = new Properties();

		InputStream is = null;
		try {
			is = PersonManagementServiceBean.class
					.getResourceAsStream("/ldap-pm.properties");
			if (is == null) {
				throw new LdapException("Could not find ldap-pm.properties file on the classpath");
			}
			try {
				properties.load(is);
			} finally {
				is.close();
			}

		} catch (IOException e) {
			throw new LdapException("Could not load ldap-pm.properties file", e);
		} finally {
			if (is != null) {
				try {
					is.close();
				} catch (IOException e) {
					// ignore
				}
			}
		}

		settings = new LdapSettings(properties);

		// create connection

		connection = new LDAPConnection();
		try {
			connection.connect(settings.getHost(), settings.getPort());
			connection.bind(LDAPConnection.LDAP_V3, settings.getServiceLogin(),
					settings.getServicePassword().getBytes(StandardCharsets.UTF_8));
		} catch (LDAPException e) {
			throw new LdapException("Failed to initialize connection to LDAP", e);
		}
	}

	@PreDestroy
	private void disconnect() {
		try {
			connection.disconnect();
			connection = null;
		} catch (LDAPException e) {
			throw new LdapException("Failed to disconnect from LDAP server", e);
		}
	}

	/**
	 * Creates a new LDAP session.
	 *
	 * @return a new LDAP session
	 */
	protected LdapSession createSession() {
		return new LdapSession(connection, settings, roleDao, securityRoleDao);
	}
}
