/*
 *********************************************************************

 $id$


 Copyright (c) 2007-2016 Whitestein Technologies AG,
 Riedstrasse 13, CH-6330 Cham, Switzerland.
 All rights reserved.

 This software is confidential and proprietary information of
 Whitestein Technologies AG.
 You shall not disclose this confidential information and shall use
 it only in accordance with the terms of the license agreement you
 entered into with Whitestein Technologies AG.
 The use of this file in source or binary form requires a written
 license from Whitestein Technologies AG.
 *********************************************************************
 */
package com.whitestein.lsps.pm.ldap;

import java.io.Serializable;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.novell.ldap.LDAPAttribute;
import com.novell.ldap.LDAPConnection;
import com.novell.ldap.LDAPEntry;
import com.novell.ldap.LDAPException;
import com.novell.ldap.LDAPModification;
import com.novell.ldap.LDAPSearchResults;
import com.novell.ldap.util.Base64;
import com.whitestein.lsps.common.ByteData;
import com.whitestein.lsps.common.Serialization;
import com.whitestein.lsps.orgstructure.entity.Role;
import com.whitestein.lsps.orgstructure.entity.SecurityRole;
import com.whitestein.lsps.os.PersonEntity;
import com.whitestein.lsps.os.dto.RoleUnitName;
import com.whitestein.lsps.os.exception.PersonNotFoundException;

/**
 * <p>This class represents a person entity stored in LDAP. Instances of this
 * class should be created exclusively through {@link LdapSession} api and
 * should be used only during the time the LSPS session is valid.</p>
 *
 *  @author jsa
 *  @see LdapSession
 */
class LdapPerson implements PersonEntity {

	private final LdapSession session;
	private final LDAPEntry entry;
	private final LdapSettings settings;

	private final String id;
	private String login;
	private String firstName;
	private String lastName;
	private String email;
	private String phone;
	private Map<String, String> details;
	private boolean substitutionActive;
	private String[] securityRoles;
	private String[] namesOfSubstitutes;

	private Set<String> rights;
	private Set<RoleUnitName> roleUnitNames;
	private Set<LdapPerson> substitutes;
	private Set<LdapPerson> substituted;
	private Map<String, ByteData> settingsMap;

	/**
	 * <p>Creates a new instance of this person.</p>
	 *
	 * <p><b>Important:</b> Do not call this constructor directly. Use
	 * {@link LdapSession} api to obtain instances of this class.</p>
	 *
	 * @param entry an LDAP entry representing the person
	 * @param session the session which is creating this person instance
	 *
	 * @see LdapSession
	 */
	LdapPerson(LDAPEntry entry, LdapSession session) {
		this.session = session;
		this.settings = session.settings;
		this.entry = entry;

		this.id = getSingleValue(entry, settings.getIdAttr());
		this.login = getSingleValue(entry, settings.getLoginAttr());
		this.firstName = getSingleValue(entry, settings.getFirstNameAttr());
		this.lastName = getSingleValue(entry, settings.getLastNameAttr());
		this.email = getSingleValue(entry, settings.getEmailAttr());
		this.phone = getSingleValue(entry, settings.getPhoneAttr());

		this.details = new HashMap<>();
		for (String attrName : settings.getPropertiesAttributes().split(",")) {
			details.put(attrName, getSingleValue(entry, attrName));
		}

		String value = getSingleValue(entry, settings.getSubstitutionActiveAttr());
		this.substitutionActive = Boolean.parseBoolean(value);

		this.securityRoles = getVectorValue(entry, settings.getSecurityRoleAttr());
		this.namesOfSubstitutes = getVectorValue(entry, settings.getSubstitutesAttr());
	}

	private String getSingleValue(LDAPEntry entry, String attribute) {
		LDAPAttribute attr = entry.getAttribute(attribute);

		if (attr == null) {
			return null;
		}

		return attr.getStringValue();
	}

	private String[] getVectorValue(LDAPEntry entry, String attribute) {
		LDAPAttribute attr = entry.getAttribute(attribute);

		if (attr == null) {
			return new String[0];
		}

		return attr.getStringValueArray();
	}

	@Override
	public String getId() {
		return id;
	}

	@Override
	public String getLoginName() {
		return login;
	}

	/**
	 * Sets the user's new login name.
	 *
	 * @param loginName the new login name value
	 */
	public void setLoginName(String loginName) {
		this.login = loginName;
		updateAttribute(settings.getLoginAttr(), loginName);
	}

	@Override
	public String getFirstName() {
		return firstName;
	}

	/**
	 * Sets the first name.
	 *
	 * @param firstName the first name to set.
	 */
	public void setFirstName(String firstName) {
		this.firstName = firstName;
		updateAttribute(settings.getFirstNameAttr(), firstName);
	}

	@Override
	public String getLastName() {
		return lastName;
	}

	/**
	 * Sets the last name.
	 *
	 * @param lastName the last name to set.
	 */
	public void setLastName(String lastName) {
		this.lastName = lastName;
		updateAttribute(settings.getLastNameAttr(), lastName);
	}

	@Override
	public String getEmail() {
		return email;
	}

	/**
	 * Sets the email attribute.
	 *
	 * @param email the email to set.
	 */
	public void setEmail(String email) {
		this.email = email;
		updateAttribute(settings.getEmailAttr(), email);
	}

	@Override
	public String getPhone() {
		return phone;
	}

	/**
	 * Sets the phone attribute.
	 *
	 * @param phone the phone to set.
	 */
	public void setPhone(String phone) {
		this.phone = phone;
		updateAttribute(settings.getPhoneAttr(), phone);
	}

	@Override
	public boolean isActive() {
		return true;
	}

	@Override
	public Long getEntityVersion() {
		return 0L;
	}

	@Override
	public Map<String, String> getDetails() {
		return Collections.unmodifiableMap(details);
	}

	/**
	 * Sets a detail
	 * @param key
	 * @param value
	 */
	public void setDetail(String key, String value) {
		updateAttribute(key, value);
	}

	public void removeDetail(String key) {
		updateAttribute(key);
	}

	@Override
	public boolean isSubstitutionActive() {
		return substitutionActive;
	}

	public void setSubstitutionActive(boolean active) {
		this.substitutionActive = active;
		updateAttribute(settings.getSubstitutionActiveAttr(), String.valueOf(active).toUpperCase());
	}

	@Override
	public Set<LdapPerson> getDirectSubstitutes() {
		if (substitutes == null) {
			substitutes = new HashSet<>();

			try {
				for (String personId : namesOfSubstitutes) {
					substitutes.add(session.loadPerson(personId));
				}
			} catch (PersonNotFoundException e) {
				throw new LdapException("Resolving relation 'DirectSubstitutes' failed.", e);
			}
		}

		return new HashSet<>(substitutes);
	}

	public void setDirectSubstitutes(Collection<String> personIds) {

		String[] array = personIds.toArray(new String[personIds.size()]);
		updateAttribute(settings.getSubstitutesAttr(), array);
		namesOfSubstitutes = array;

		// clean attributes which depends on the substitutes
		substitutes = null;
		substituted = null;
	}

	@Override
	public Set<LdapPerson> getDirectlySubstituted() {

		if (substituted == null) {
			substituted = new HashSet<>();
			StringBuilder query = new StringBuilder("(&");
			query.append(settings.getUserGenericQueryCondition());
			query.append("(")
					.append(settings.getSubstitutesAttr())
					.append("=")
					.append(getId())
					.append(")");
			query.append(")");

			try {
				LDAPSearchResults results = session.connection.search(settings.getBase(),
						LDAPConnection.SCOPE_ONE, query.toString(), null, false);
				while (results.hasMore()) {
					LDAPEntry entry = results.next();
					String personId = getSingleValue(entry, settings.getIdAttr());

					substituted.add(session.loadPerson(personId));
				}
			} catch (LDAPException | PersonNotFoundException e) {
				throw new LdapException("Resolving relation 'DirectlySubstituted' failed.", e);
			}
		}

		return substituted;
	}

	@Override
	public Set<LdapPerson> getActiveSubstitutes() {
		Set<LdapPerson> result = new HashSet<>();
		calcActiveSubstitutes(this, result);
		// we don't want this LdapPerson in the result even in a case of cyclic
		// substitutions
		result.remove(this);
		return result;
	}

	private void calcActiveSubstitutes(LdapPerson person, Set<LdapPerson> result) {
		if (person.isSubstitutionActive()) {
			for (LdapPerson sub : person.getDirectSubstitutes()) {
				if (!result.contains(sub)) {
					result.add(sub);
					calcActiveSubstitutes(sub, result);
				}
			}
		}
	}

	@Override
	public Set<LdapPerson> getActivelySubstituted() {
		Set<LdapPerson> allSubstitutedPersons = new HashSet<>();
		searchForSubstitutedBy(this, allSubstitutedPersons);
		return allSubstitutedPersons;
	}

	private void searchForSubstitutedBy(LdapPerson person, Set<LdapPerson> allSubstitutedPersons) {
		for (LdapPerson substituted : person.getDirectlySubstituted()) {
			if (substituted != this && substituted.isSubstitutionActive()
					&& !allSubstitutedPersons.contains(substituted)) {
				allSubstitutedPersons.add(substituted);
				searchForSubstitutedBy(substituted, allSubstitutedPersons);
			}
		}
	}

	@Override
	public Set<RoleUnitName> getRoleUnitNames() {
		if (roleUnitNames == null) {
			roleUnitNames = new HashSet<>();
			Collection<Role> roles = session.roleDao.findForPerson(getId());
			for (Role role : roles) {
				roleUnitNames.add(role.getFullName());
			}
		}

		return new HashSet<>(roleUnitNames);
	}

	@Override
	public Set<String> getRights() {

		if (rights == null) {
			rights = new HashSet<>();

			for (String role : securityRoles) {
				SecurityRole securityRole = session.securityRoleDao.find(role);
				if (securityRole != null) {
					rights.addAll(securityRole.getRights());
				} else {
					//ignore
				}
			}
		}

		return new HashSet<>(rights);
	}

	/**
	 * Gets the assigned security roles.
	 *
	 * @return an array of assigned security roles.
	 */
	public Collection<String> getSecurityRoles() {
		return Collections.unmodifiableList(Arrays.asList(securityRoles));
	}

	/**
	 * Updates the security role assignments.
	 *
	 * @param securityRoles the only roles this person should have
	 */
	public void setSecurityRoleAssignments(String... securityRoles) {
		updateAttribute(settings.getSecurityRoleAttr(), securityRoles);
		this.securityRoles = Arrays.copyOf(securityRoles, securityRoles.length);

		// do cleanup to reload obsolete data
		rights = null;
	}

	@SuppressWarnings("unchecked")
	public Map<String, ByteData> getSettings() {
		if (settingsMap == null) {
			String base64string = getSingleValue(entry, settings.getSettingsAttr());
			if (base64string != null && base64string.trim().length() > 0) {
				settingsMap = (Map<String, ByteData>) Serialization.deserialize(Base64.decode(base64string));
			} else {
				settingsMap = new HashMap<>();
			}
		}

		return Collections.unmodifiableMap(settingsMap);
	}

	/**
	 * Updates the setting map.
	 * @param settingsMap
	 */
	public void setSettings(Map<String, ByteData> settingsMap) {

		String base64string = Base64.encode(Serialization.serialize((Serializable) settingsMap));
		LDAPAttribute attributeSettings = new LDAPAttribute(settings.getSettingsAttr(), base64string);
		try {
			session.connection.modify(entry.getDN(), new LDAPModification(LDAPModification.REPLACE, attributeSettings));
		} catch (LDAPException e) {
			throw new LdapException("Updating the attribute 'setting' failed.", e);
		}

		this.settingsMap = new HashMap<>(settingsMap);
	}

	/**
	 * Gets a specific setting value.
	 *
	 * @param name the setting name
	 *
	 * @return the setting value.
	 */
	public ByteData getSetting(String name) {
		return getSettings().get(name);
	}

	/**
	 * Updates the user's password in LDAP.
	 *
	 * @param password the new password value
	 */
	public void updatePassword(String password) {
		updateAttribute(settings.getPasswordAttr(), password);
	}

	private void updateAttribute(String attrName, String value) {
		if (value != null && value.trim().length() > 0) {
			updateAttribute(attrName, new String[] { value });
		} else {
			updateAttribute(attrName); // this actually removes the attribute
		}
	}

	private void updateAttribute(String attrName, String... content) {
		try {
			LDAPAttribute updatedAttribute = new LDAPAttribute(attrName, content);
			session.connection.modify(entry.getDN(), new LDAPModification(LDAPModification.REPLACE, updatedAttribute));
		} catch (LDAPException e) {
			throw new LdapException("Unable to perform LDAP operation", e);
		}
	}

}
