/*
 *********************************************************************

 $id$


 Copyright (c) 2007-2016 Whitestein Technologies AG,
 Riedstrasse 13, CH-6330 Cham, Switzerland.
 All rights reserved.

 This software is confidential and proprietary information of
 Whitestein Technologies AG.
 You shall not disclose this confidential information and shall use
 it only in accordance with the terms of the license agreement you
 entered into with Whitestein Technologies AG.
 The use of this file in source or binary form requires a written
 license from Whitestein Technologies AG.
 *********************************************************************
 */
package com.whitestein.lsps.pm.ldap;

import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import com.novell.ldap.LDAPAttribute;
import com.novell.ldap.LDAPAttributeSet;
import com.novell.ldap.LDAPConnection;
import com.novell.ldap.LDAPEntry;
import com.novell.ldap.LDAPException;
import com.novell.ldap.LDAPObjectClassSchema;
import com.novell.ldap.LDAPSchema;
import com.novell.ldap.LDAPSearchResults;
import com.whitestein.lsps.orgstructure.dao.RoleDao;
import com.whitestein.lsps.orgstructure.dao.SecurityRoleDao;
import com.whitestein.lsps.os.exception.PersonAlreadyExistsException;
import com.whitestein.lsps.os.exception.PersonNotFoundException;

/**
 * This class represents a session with an LDAP server. A lifetime of an LDAP
 * session should be only as long as it is necessary to handle a single user
 * request. Once the request is handled the session should be thrown away.
 *
 * @author jsa
 *
 */
class LdapSession {

	/**
	 * LDAP connection.
	 */
	public final LDAPConnection connection;
	/**
	 * Settings of this LDAP session.
	 */
	public final LdapSettings settings;
	/**
	 * DAO for Role entity.
	 */
	public final RoleDao roleDao;
	/**
	 * DAO for security role entity.
	 */
	public final SecurityRoleDao securityRoleDao;

	private final Map<String, LdapPerson> persons;

	/**
	 * @param connection
	 * @param settings
	 * @param roleDao
	 * @param securityRoleDao
	 */
	LdapSession(LDAPConnection connection, LdapSettings settings, RoleDao roleDao, SecurityRoleDao securityRoleDao) {
		super();
		this.connection = connection;
		this.settings = settings;
		this.roleDao = roleDao;
		this.securityRoleDao = securityRoleDao;

		this.persons = new HashMap<>();
	}

	/**
	 * Loads a person from LDAP. If the person has been already loaded by this
	 * session then it returns the loaded value.
	 *
	 * @param personId the person identifier
	 *
	 * @return the loaded person.
	 * @throws PersonNotFoundException if no person with the given person
	 * 			identifier is found
	 */
	public LdapPerson loadPerson(String personId) throws PersonNotFoundException {
		LdapPerson person = persons.get(personId);
		if (person == null) {

			String query = MessageFormat.format(settings.getUserQueryById(), personId);

			LDAPSearchResults searchResults;
			try {
				searchResults = connection.search(settings.getBase(),
						LDAPConnection.SCOPE_ONE, query, null, false);

				if (!searchResults.hasMore()) {
					throw new PersonNotFoundException(personId);
				}

				LDAPEntry entry = searchResults.next();

				if (searchResults.hasMore()) {
					throw new LdapException("More than one person found for the ID: " + personId);
				}

				person = new LdapPerson(entry, this);
				persons.put(personId, person);
			} catch (LDAPException e) {
				throw new LdapException("Creating new LdapPerson entity failed.", e);
			}

		}

		return person;
	}

	/**
	 * Creates a new LDAP entry representing a user.
	 *
	 * @param personId the identifier of the user.
	 *
	 * @return the created person
	 * @throws PersonAlreadyExistsException
	 */
	public LdapPerson createPerson(String personId) throws PersonAlreadyExistsException {
		String[] objectClasses;
		try {
			List<String> hierarchy = new ArrayList<>();
			fetchObjectClassHierarchy(settings.getUserObjectClass(), hierarchy);
			objectClasses = hierarchy.toArray(new String[hierarchy.size()]);
		} catch (LDAPException e) {
			throw new LdapException("Retrieving objectClass hierarchy for the objectClass='" + settings.getUserObjectClass() + "' failed.", e);
		}

		// temporarily fill name related attributes with the personId
		LDAPAttributeSet attributeSet = new LDAPAttributeSet();
		attributeSet.add(new LDAPAttribute("objectClass", objectClasses));
		attributeSet.add(new LDAPAttribute(settings.getIdAttr(), personId));
		attributeSet.add(new LDAPAttribute(settings.getLoginAttr(), personId));
		attributeSet.add(new LDAPAttribute("cn", personId));
		attributeSet.add(new LDAPAttribute(settings.getLastNameAttr(), personId));

		LDAPEntry entry = new LDAPEntry(settings.getIdAttr() + "=" + personId + "," + settings.getBase(), attributeSet);
		try {
			connection.add(entry);
		} catch (LDAPException e) {
			throw new LdapException("Adding LDAP entry failed.", e);
		}

		try {
			return loadPerson(personId);
		} catch (PersonNotFoundException e) {
			// this should never happen as the person has just been created
			throw new RuntimeException(e);
		}
	}

	/**
	 * Walks the objectClass hierarchy in the bottom-up direction and collects
	 * all the object classes found on the path.
	 *
	 * @param objectClass the object class to start with
	 *
	 * @param result output parameter used to store found object classes
	 *
	 * @throws LDAPException
	 */
	private void fetchObjectClassHierarchy(String objectClass, List<String> result) throws LDAPException {
		result.add(objectClass);

		LDAPSchema schema = connection.fetchSchema(connection.getSchemaDN());
		LDAPObjectClassSchema objectClassSchema = schema.getObjectClassSchema(objectClass);
		String[] superiors = objectClassSchema.getSuperiors();

		if (superiors != null) {
			for (String superior : superiors) {
				if (!result.contains(superior)) {
					fetchObjectClassHierarchy(superior, result);
				}
			}
		}
	}

	/**
	 * Converts an LDAPEntry representing a person to {@link LdapPerson}.
	 *
	 * @param entry the entry representing a person.
	 *
	 * @return an LDAP person entity.
	 */
	private LdapPerson toPerson(LDAPEntry entry) {
		if (!entry.getDN().endsWith(settings.getBase())) {
			throw new IllegalArgumentException("LDAPEntry does not represent an LSPS person.");
		}

		LDAPAttribute idAttr = entry.getAttribute(settings.getIdAttr());
		assert idAttr != null;

		final String personId = idAttr.getStringValue();
		LdapPerson person = persons.get(personId);
		if (person == null) {
			person = new LdapPerson(entry, this);
			persons.put(personId, person);
		}

		return person;
	}

	/**
	 * Executes a person query and returns the result as a list of found persons.
	 *
	 * @param query the person query to execute.
	 *
	 * @return a list of found persons
	 */
	public List<LdapPerson> searchPersons(String query) {
		try {
			LDAPSearchResults searchResults = connection.search(settings.getBase(),
					LDAPConnection.SCOPE_ONE, query, null, false);

			List<LdapPerson> result = new ArrayList<>();

			while (searchResults.hasMore()) {
				result.add(toPerson(searchResults.next()));
			}
			return result;
		} catch (LDAPException e) {
			throw new LdapException("Unable to perform LDAP operation", e);
		}
	}
}
