/*
 *********************************************************************

 $Id$


 Copyright (c) 2007-2014 Whitestein Technologies AG,
 Riedstrasse 13, CH-6330 Cham, Switzerland.
 All rights reserved.

 This software is confidential and proprietary information of
 Whitestein Technologies AG.
 You shall not disclose this confidential information and shall use
 it only in accordance with the terms of the license agreement you
 entered into with Whitestein Technologies AG.
 The use of this file in source or binary form requires a written
 license from Whitestein Technologies AG.
 *********************************************************************
 */
package com.whitestein.lsps.pm.ldap;

import java.io.IOException;
import java.nio.CharBuffer;
import java.nio.charset.CharacterCodingException;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.StandardCharsets;
import java.nio.charset.UnmappableCharacterException;
import java.text.MessageFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Objects;
import java.util.Set;

import javax.annotation.security.PermitAll;
import javax.ejb.Stateless;
import javax.xml.bind.JAXBException;

import com.novell.ldap.LDAPConnection;
import com.novell.ldap.LDAPEntry;
import com.novell.ldap.LDAPException;
import com.novell.ldap.LDAPSearchResults;
import com.whitestein.lsps.common.ByteData;
import com.whitestein.lsps.common.RightsAllowed;
import com.whitestein.lsps.common.collections.Diff;
import com.whitestein.lsps.orgstructure.PersonConverter;
import com.whitestein.lsps.orgstructure.SecurityRoleConverter;
import com.whitestein.lsps.orgstructure.entity.Role;
import com.whitestein.lsps.orgstructure.entity.RoleName;
import com.whitestein.lsps.orgstructure.entity.SecurityRole;
import com.whitestein.lsps.os.OsMarshaller;
import com.whitestein.lsps.os.OsRights;
import com.whitestein.lsps.os.PersonAttribute;
import com.whitestein.lsps.os.PersonManagementService;
import com.whitestein.lsps.os.PersonService;
import com.whitestein.lsps.os.dto.Person;
import com.whitestein.lsps.os.dto.PersonCriteria;
import com.whitestein.lsps.os.dto.PersonData;
import com.whitestein.lsps.os.dto.PersonList;
import com.whitestein.lsps.os.dto.PersonPicture;
import com.whitestein.lsps.os.dto.RoleUnitName;
import com.whitestein.lsps.os.dto.RoleUnitNameList;
import com.whitestein.lsps.os.ejb.PersonManagementServiceLocal;
import com.whitestein.lsps.os.ejb.PersonManagementServiceRemote;
import com.whitestein.lsps.os.exception.InvalidPasswordException;
import com.whitestein.lsps.os.exception.PersonAlreadyExistsException;
import com.whitestein.lsps.os.exception.PersonNotFoundException;
import com.whitestein.lsps.os.exception.PersonPictureException;
import com.whitestein.lsps.os.exception.PersonValidationException;

/**
 * Implementation of {@link PersonManagementService} and {@link PersonService}
 * for LDAP based user management.
 *
 * @author jsa
 *
 */
@Stateless
@PermitAll
public class PersonManagementServiceBean extends AbstractLdapBean implements PersonManagementServiceLocal, PersonManagementServiceRemote {

	/**
	 * The name of the key to the settingsMap which holds the person picture
	 * value.
	 */
	private static final String PERSON_PICTURE = "PersonPicture";

	@Override
	@RightsAllowed(OsRights.READ_PERSON)
	public Person getPerson(String id) {
		LdapSession session = createSession();
		try {
			LdapPerson person = session.loadPerson(id);
			return PersonConverter.FIRST_LEVEL_SUB.convert(person);
		} catch (PersonNotFoundException e) {
			return null;
		}
	}

	@Override
	@RightsAllowed(OsRights.READ_PERSON)
	public Person getPersonByLogin(String login) {
		LdapSession session = createSession();
		List<LdapPerson> matchedPersons = session.searchPersons(MessageFormat.format(settings.getUserQueryByLogin(), login));

		if (matchedPersons.size() == 0) {
			return null;
		} else if (matchedPersons.size() == 1) {
			return PersonConverter.FIRST_LEVEL_SUB.convert(matchedPersons.get(0));
		} else {
			throw new LdapException("More than one person found for the login: " + login);
		}
	}

	@Override
	public boolean checkActivePersonWithLoginExists(String login) {
		LdapSession session = createSession();
		List<LdapPerson> matchedPersons = session.searchPersons(MessageFormat.format(settings.getUserQueryByLogin(), login));

		if (matchedPersons.size() == 0) {
			return false;
		} else if (matchedPersons.size() == 1) {
			LdapPerson person = matchedPersons.get(0);
			return person.isActive();
		} else {
			throw new LdapException("More than one person found for the login: " + login);
		}
	}

	@Override
	@RightsAllowed(OsRights.READ_PERSON)
	public PersonList findPersons(PersonCriteria criteria) {

		int counter = 0;
		List<String> params = new ArrayList<>();
		StringBuilder query = new StringBuilder(settings.getUserGenericQueryCondition());

		if (criteria.getEmailPattern() != null) {
			query.append("({" + counter++ + "}={" + counter++ + "})");
			params.add(settings.getEmailAttr());
			params.add(criteria.getEmailPattern());

		}

		if (criteria.getFirstNamePattern() != null) {
			query.append("({" + counter++ + "}={" + counter++ + "})");
			params.add(settings.getFirstNameAttr());
			params.add(criteria.getFirstNamePattern());
		}

		if (criteria.getLastNamePattern() != null) {
			query.append("({" + counter++ + "}={" + counter++ + "})");
			params.add(settings.getLastNameAttr());
			params.add(criteria.getLastNamePattern());
		}

		if (criteria.getLoginNamePattern() != null) {
			query.append("({" + counter++ + "}={" + counter++ + "})");
			params.add(settings.getLoginAttr());
			params.add(criteria.getLoginNamePattern());
		}

		if (criteria.getPhonePattern() != null) {
			query.append("({" + counter++ + "}={" + counter++ + "})");
			params.add(settings.getPhoneAttr());
			params.add(criteria.getPhonePattern());
		}

		// paging is not supported

		LdapSession session = createSession();
		List<LdapPerson> result = session.searchPersons(MessageFormat.format("(&" + query + ")",
				(Object[]) params.toArray(new String[params.size()])));

		final RoleUnitName criteriaRoleUnitName = criteria.getInRole();

		if (criteriaRoleUnitName != null) {
			Iterator<LdapPerson> iterator = result.iterator();
			while (iterator.hasNext()) {
				LdapPerson person = iterator.next();

				boolean isInRole = false;
				for (RoleUnitName personRoleUnitName : person.getRoleUnitNames()) {
					if (criteriaRoleUnitName.matches(personRoleUnitName)) {
						isInRole = true;
						break;
					}
				}

				if (!isInRole) {
					iterator.remove();
				}
			}
		}

		PersonConverter conv = criteria.isIncludeSubstitution() ? PersonConverter.FIRST_LEVEL_SUB
				: PersonConverter.NO_SUB;

		return new PersonList(conv.convert(result), result.size());
	}

	@Override
	@RightsAllowed(OsRights.READ_PERSON)
	public ByteData getSetting(String personId, String settingName) throws PersonNotFoundException {
		LdapPerson person = createSession().loadPerson(personId);
		return person.getSetting(settingName);
	}

	@Override
	@RightsAllowed(OsRights.READ_PERSON)
	public void setSetting(String personId, String settingName, ByteData settingContent) throws PersonNotFoundException {
		LdapPerson person = createSession().loadPerson(personId);
		Map<String, ByteData> settings = new HashMap<>(person.getSettings());
		settings.put(settingName, settingContent);
		person.setSettings(settings);
	}

	@Override
	@RightsAllowed(OsRights.READ_PERSON)
	public void deleteSetting(String personId, String settingName)
			throws PersonNotFoundException {
		LdapPerson person = createSession().loadPerson(personId);
		Map<String, ByteData> settings = new HashMap<>(person.getSettings());
		settings.remove(settingName);
		person.setSettings(settings);
	}

	@Override
	@RightsAllowed(OsRights.READ_PERSON)
	public boolean hasPersonAnyRight(String loginName, List<String> rights) {
		Person person = getPersonByLogin(loginName);
		if ((person == null) || (!person.isActive())) {
			return false;
		}
		if ((rights == null) || rights.isEmpty()) {
			return true;
		}
		Set<String> personRights = person.getRights();
		for (String neededRight : rights) {
			if (personRights.contains(neededRight)) {
				return true;
			}
		}
		return false;
	}

	protected LDAPEntry getEntryForUser(String personId) throws LDAPException, PersonNotFoundException {
		LDAPSearchResults searchResults = connection.search(settings.getBase(),
				LDAPConnection.SCOPE_ONE, MessageFormat.format(settings.getUserQueryById(), personId), null, false);

		if (searchResults.hasMore()) {
			return searchResults.next();
		}
		throw new PersonNotFoundException(personId);
	}

	@Override
	public Collection<com.whitestein.lsps.os.dto.SecurityRole> getSecurityRoles(String personId) throws PersonNotFoundException {
		LdapPerson person = createSession().loadPerson(personId);
		Collection<SecurityRole> roles = new ArrayList<>();

		for (String role : person.getSecurityRoles()) {
			SecurityRole securityRole = securityRoleDao.find(role);
			if (securityRole != null) {
				roles.add(securityRole);
			}
		}
		return SecurityRoleConverter.convert(roles);
	}

	@Override
	public Person assignSecurityRolesToPerson(String personId, List<String> securityRoleNames) throws PersonNotFoundException {
		String[] rolesArray = securityRoleNames.toArray(new String[securityRoleNames.size()]);
		LdapPerson ldapPerson = createSession().loadPerson(personId);
		ldapPerson.setSecurityRoleAssignments(rolesArray);
		return PersonConverter.FIRST_LEVEL_SUB.convert(ldapPerson);
	}

	@Override
	public Person updatePersonLogin(String personId, String newLogin)
			throws PersonNotFoundException {
		LdapPerson ldapPerson = createSession().loadPerson(personId);
		ldapPerson.setLoginName(newLogin);
		return PersonConverter.FIRST_LEVEL_SUB.convert(ldapPerson);
	}

	@Override
	public Person updatePersonPassword(String personId, String newPassword)
			throws PersonNotFoundException {
		LdapPerson ldapPerson = createSession().loadPerson(personId);
		ldapPerson.updatePassword(newPassword);
		return PersonConverter.FIRST_LEVEL_SUB.convert(ldapPerson);
	}

	@Override
	public Person updatePerson(Person person) throws PersonNotFoundException {

		LdapSession session = createSession();
		LdapPerson ldapPerson = session.loadPerson(person.getId());

		updateLdapPersonAttributes(person, ldapPerson);
		return person;
	}

	private void updateLdapPersonAttributes(Person person, LdapPerson ldapPerson) {
		// In our LDAP the surname is a mandatory attribute. Fill it with
		// login if the clien does not provide one
		ldapPerson.setLastName(person.getLastName());
		ldapPerson.setFirstName(person.getFirstName());
		ldapPerson.setPhone(person.getPhone());
		ldapPerson.setEmail(person.getEmail());
		ldapPerson.setSubstitutionActive(person.isSubstitutionActive());
		ldapPerson.setDirectSubstitutes(getPersonIds(person.getDirectSubstitutes()));

		for (String attrName : settings.getPropertiesAttributes().split(",")) {
			ldapPerson.setDetail(attrName, person.getDetail(attrName));
		}

		Diff<RoleUnitName> diff = Diff.of(ldapPerson.getRoleUnitNames(), person.getRoleUnitNames());

		if (!diff.added().isEmpty() || !diff.removed().isEmpty()) {
			Collection<Role> roles = roleDao.findForPerson(person.getId());
			for (Role role : roles) {
				if (diff.removed().contains(role.getFullName())) {
					roleDao.remove(role);
				}
			}

			for (RoleUnitName roleUnitName : diff.added()) {
				RoleName roleName = roleNameDao.findById(roleUnitName.getModuleName(), roleUnitName.getLocalName());
				Role role = new Role(roleName, roleUnitName);
				role.setPersonId(person.getId());
				roleDao.persist(role);
			}
		}
	}

	private Collection<String> getPersonIds(Collection<Person> persons) {
		Set<String> ids = new LinkedHashSet<>(persons.size());
		for (Person person : persons) {
			ids.add(person.getId());
		}

		return ids;
	}

	@Override
	public Person addPerson(Person person) throws PersonAlreadyExistsException, PersonValidationException {
		Objects.requireNonNull(person, "Parameter 'person' must not be null.");

		validatePerson(person);

		LdapSession session = createSession();
		List<LdapPerson> matchedPersons = session.searchPersons(MessageFormat.format(settings.getUserQueryByLogin(), person.getLoginName()));

		if (matchedPersons.size() != 0) {
			throw new PersonAlreadyExistsException(person.getLoginName());
		}

		LdapPerson ldapPerson = session.createPerson(person.getLoginName());
		updateLdapPersonAttributes(person, ldapPerson);

		return PersonConverter.FIRST_LEVEL_SUB.convert(ldapPerson);
	}

	@Override
	public RoleUnitNameList findRoles() {
		Collection<Role> roles = roleDao.findAll();
		Set<RoleUnitName> names = new HashSet<>();
		for (Role role : roles) {
			names.add(role.getFullName());
		}
		return new RoleUnitNameList(names, names.size());
	}

	@Override
	public boolean isReadOnly() {
		return settings.isReadonly();
	}

	@Override
	public Collection<PersonAttribute> getEditableAttributes() {
		if (settings.isReadonly()) {
			// Modeled roles are not stored in the LDAP so they can be editable
			// even if LDAP is read only.
			return Arrays.asList(PersonAttribute.MODELED_ROLES);
		} else {
			return Arrays.asList(PersonAttribute.values());
		}
	}

	@Override
	@RightsAllowed(OsRights.READ_PERSON)
	public PersonPicture getPersonPicture(String personId) throws PersonNotFoundException, PersonPictureException {
		LdapSession session = createSession();
		LdapPerson person = session.loadPerson(personId);

		ByteData byteData = person.getSetting(PERSON_PICTURE);
		if (byteData == null) {
			return null;
		}

		try {
			return (PersonPicture) new OsMarshaller().unmarshall(byteData.getData());
		} catch (JAXBException | IOException e) {
			throw new PersonPictureException("Exception to unmarshal person picture.", e);
		}
	}

	@Override
	@RightsAllowed(OsRights.READ_PERSON)
	public void setPersonPicture(String personId, PersonPicture picture) throws PersonNotFoundException, PersonPictureException {
		if (picture == null) {
			deleteSetting(personId, PERSON_PICTURE);
		}

		byte[] bs;
		try {
			bs = new OsMarshaller().marshall(picture);
		} catch (JAXBException | IOException e) {
			throw new PersonPictureException("Exception to marshall person picture", e);
		}
		setSetting(personId, PERSON_PICTURE, new ByteData(bs));

	}

	@Override
	public void validatePassword(String password) throws InvalidPasswordException {
		CharsetEncoder encoder = StandardCharsets.ISO_8859_1.newEncoder();

		if (!encoder.canEncode(password)) {
			// The JavaDoc of canEncode(...) recommends to perform the encode
			// operation to obtain more information why the string cannot be
			// encoded.
			try {
				encoder.encode(CharBuffer.wrap(password));
			} catch (UnmappableCharacterException e) {
				String errMsg = "The password is invalid. It contains unsupported characters."
						+ " Only characters from the character set "
						+ StandardCharsets.ISO_8859_1.name() + " are supported.";
				throw new InvalidPasswordException(InvalidPasswordException.ERROR_CODE_NON_ENCODABLE_ISO_8859_1, errMsg, e);
			} catch (CharacterCodingException e) {
				String errMsg = "The password is invalid. It cannot be encoded"
						+ " using the " + StandardCharsets.ISO_8859_1.name() + " encoder.";
				throw new InvalidPasswordException(InvalidPasswordException.ERROR_CODE_NON_ENCODABLE_ISO_8859_1, errMsg, e);
			}
		}
	}

	private void validatePerson(Person person) throws PersonValidationException {
		if (person.getLoginName() == null || person.getLoginName().trim().isEmpty()) {
			throw new PersonValidationException("Attribute 'login' must not be empty.");
		}

		if (person.getLastName() == null || person.getLastName().trim().isEmpty()) {
			throw new PersonValidationException("Attribute 'Last Name' must not be empty.");
		}
	}

	@Override
	public Person addPersonWithData(PersonData personData) throws PersonAlreadyExistsException,
			PersonNotFoundException, PersonValidationException {
		Person person = Objects.requireNonNull(personData.getPerson());
		validatePerson(person);

		person = addPerson(person);
		final String personId = person.getId();

		String newPassword = personData.getPassword();
		if (newPassword != null) {
			updatePersonPassword(personId, newPassword);
		}

		List<String> securityRoleNames = personData.getSecurityRoleNames();

		if (securityRoleNames != null) {
			assignSecurityRolesToPerson(personId, securityRoleNames);
		}

		return person;
	}

	@Override
	public Person updatePersonWithData(PersonData personData) throws PersonNotFoundException {
		Person person = Objects.requireNonNull(personData.getPerson());

		final String personId = person.getId();
		if (personId == null) {
			throw new IllegalArgumentException("The provided person does not have an ID.");
		}

		person = updatePerson(person);

		String newPassword = personData.getPassword();
		if (newPassword != null) {
			updatePersonPassword(personId, newPassword);
		}

		List<String> securityRoleNames = personData.getSecurityRoleNames();

		if (securityRoleNames != null) {
			assignSecurityRolesToPerson(personId, securityRoleNames);
		}

		return person;
	}

}
